from typing import Any, Callable, Tuple

from plover import system
from plover.engine import StenoEngine
from plover.oslayer.config import PLUGINS_PLATFORM
from plover.gui_qt.tool import Tool
from plover.steno import Stroke

from PyQt5.QtWidgets import QAction, QHBoxLayout, QGraphicsView
from PyQt5.QtGui import QKeySequence, QMouseEvent, QColor
from PyQt5.QtCore import Qt, QRect, QSettings

from plover_svg_layout_display.resources_rc import *
from plover_svg_layout_display.config_ui import ConfigUI
from plover_svg_layout_display.layout_config import CONFIG_ITEMS, CONFIG_TYPES, SYSTEM_PREFIX, LayoutConfig
from plover_svg_layout_display.svg_widget import LayoutWidget
from plover_svg_layout_display.qt_utils import load_qt_text



STYLESHEET = "border:0px; background:transparent;"
DEFAULT_SVG = ":/svgld/en_layout.svg"
DEFAULT_PY = ":/svgld/en_convert.py"


class SVGLayoutDisplayTool(Tool):
    TITLE = "SVG Layout Display"
    ICON = ":/svgld/icon.svg"
    ROLE = "svgld"

    def __init__(self, engine: StenoEngine) -> None:
        super().__init__(engine)
        self.setObjectName("svgld")
        engine.signal_connect("stroked", self.on_stroke)
        engine.signal_connect("config_changed", self.on_config_changed)

        self.system_name = system.NAME
        self.repaint_offset = False
        self.convert_stroke = None

        self.config = LayoutConfig()
        self.restore_state()
        self.setup_actions()
        self.setup_trans()
        self.setup_layout()
        self.reload_config()

        self.finished.connect(self.save_state)

    def _restore_state(self, settings: QSettings) -> None:
        # Cross system settings
        for field_name in CONFIG_ITEMS.keys():
            if settings.contains(field_name) and not field_name.startswith(SYSTEM_PREFIX):
                setattr(
                    self.config, 
                    field_name, 
                    settings.value(field_name, type=CONFIG_TYPES[field_name])
                )

            elif (
                field_name == "force_repaint"
                and PLUGINS_PLATFORM is not None and PLUGINS_PLATFORM == "mac"
            ):
                self.config.force_repaint = True
        
        # System specific settings
        for field_name in settings.allKeys():
            if "/" in field_name:
                sys_name, sys_field = field_name.split("/", 1)
                if sys_field in CONFIG_ITEMS.keys():
                    if sys_name not in self.config.system_map:
                        self.config.system_map[sys_name] = dict()
                    
                    self.config.system_map[sys_name][sys_field] = settings.value(
                        field_name,
                        type=CONFIG_TYPES[sys_field]
                    )

    def _save_state(self, settings: QSettings) -> None:
        for key, value in self.config.get_values():
            settings.setValue(key, value)
        
        for sys_name, sys_map in self.config.system_map.items():
            for key, value in sys_map.items():
                settings.setValue(sys_name + "/" + key, value)
    
    def view_mouse_move(self, event: QMouseEvent) -> None:
        if event.buttons() & Qt.LeftButton:
            self.move(event.globalPos() - self.drag_position)

    def view_mouse_press(self, event: QMouseEvent) -> None:
        if event.buttons() & Qt.LeftButton:
            self.drag_position = event.globalPos() - self.frameGeometry().topLeft()
    
    def setup_actions(self) -> None:
        self.close_action = QAction(self)
        self.close_action.setText("Close Display")
        self.close_action.triggered.connect(self.accept)
        self.close_action.setShortcut(QKeySequence("Ctrl+X"))
        self.addAction(self.close_action)

        self.settings_action = QAction(self)
        self.settings_action.setText("Configure Display")
        self.settings_action.triggered.connect(self.on_settings)
        self.settings_action.setShortcut(QKeySequence("Ctrl+S"))
        self.addAction(self.settings_action)
    
    def on_stroke(self, stroke_tup: Tuple[str, ...]) -> None:
        if self.convert_stroke is not None:
            prev_translations = self._engine.translator_state.prev()
            if not prev_translations:
                output = ""
            else:
                output = prev_translations[-1].english

            group_ids = self.convert_stroke(stroke_tup, output)
            self.svg_widget.update_groups(group_ids)

        self.repaint()

    def setup_trans(self) -> None:
        # For some strange reason, even though this piece of code doesn't
        # actually draw anything, the widget refuses to be transparent on
        # some systems unless this is included.
        self.trans_view = QGraphicsView(self)
        self.trans_view.setStyleSheet(STYLESHEET)

    def setup_layout(self) -> None:
        self.setWindowFlags(self.windowFlags() | Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)
        self.setAttribute(Qt.WA_TranslucentBackground)
        self.setStyleSheet("QWidget#svgld {background:transparent;}")

        self.svg_widget = LayoutWidget()

        self.layout = QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self.svg_widget)
        self.setLayout(self.layout)

        self.mouseMoveEvent = self.view_mouse_move
        self.mousePressEvent = self.view_mouse_press

        self.show()

    def on_config_changed(self, config: dict) -> None:
        if "system_name" not in config:
            return
        
        new_sys_name = config["system_name"]
        if new_sys_name == self.system_name:
            return
 
        self.reload_config()
    
    def on_settings(self) -> None:
        config_dialog = ConfigUI(self.config.copy(), self.system_name, self)
        if config_dialog.exec():
            self.config = config_dialog.temp_config
            self.reload_config()
    
    def load_py_script(self, py_path: str) -> None:
        py_text = load_qt_text(py_path)

        if not py_text.strip():
            self.convert_stroke = None
            return

        try:
            globs = {}
            exec(py_text, globs)

            convert_stroke = globs.get("convert_stroke")
            if not isinstance(convert_stroke, Callable):
                self.convert_stroke = None
            
            self.convert_stroke = convert_stroke

        except:
            self.convert_stroke = None

    def reload_config(self) -> None:
        if self.system_name in self.config.system_map:
            sys_config = self.config.system_map[self.system_name]
            if "system_svg" in sys_config:
                svg_path = sys_config["system_svg"]
                self.svg_widget.load_svg(
                    svg_path, 
                    sys_config.get("system_scale", 100)
                )

            if "system_py" in sys_config:
                py_path = sys_config["system_py"]
                self.load_py_script(py_path)
                
        else:
            if self.system_name == "English Stenotype":
                self.svg_widget.load_svg(
                    DEFAULT_SVG, 
                    sys_config.get("system_scale", 100)
                )
                self.load_py_script(DEFAULT_PY)
        
        self.on_stroke(tuple())

    def repaint_rect(self) -> QRect:
        window_rect = self.rect()
        if self.repaint_offset:
            window_rect.setWidth(window_rect.width() - 1)
        
        return window_rect

    def repaint(self) -> None:
        self.repaint_offset = not self.repaint_offset
        svg_size = self.svg_widget.svg_size

        if svg_size is not None:
            right_margin_px = 1 * self.config.force_repaint * self.repaint_offset
            self.layout.setContentsMargins(0, 0, right_margin_px, 0)
            self.setFixedWidth(svg_size.width() + right_margin_px)
            self.setFixedHeight(svg_size.height())

    # def repaint(self) -> None:
    #     self.repaint_offset = not self.repaint_offset
    #     self.setFixedWidth(self.width + self.repaint_offset * self.config.force_repaint_px)
        



from PyQt5.QtCore import QFile, QIODevice, QTextStream


def load_qt_text(text_path: str) -> str:
    try:
        file = QFile(text_path)
        text = ""
        if file.open(QIODevice.ReadOnly | QFile.Text):
            text_stream = QTextStream(file)
            text_stream.setCodec("utf-8")
            text_stream.setAutoDetectUnicode(True)

            text = text_stream.readAll()
            file.close()
        return text
    except:
        return ""



from PyQt5.QtWidgets import (
    QDialog, QWidget, QLabel, QDialogButtonBox, QGridLayout,
    QGroupBox, QCheckBox, QVBoxLayout, QLineEdit, QScrollArea,
    QFileDialog, QPushButton, QSpinBox
)
from PyQt5.QtCore import Qt

from plover_svg_layout_display.layout_config import (
    CONFIG_FILE_PARAMS, SYSTEM_NAME_PLACEHOLDER, SYSTEM_PREFIX, 
    LayoutConfig, CONFIG_NAMES, CONFIG_ORDER, CONFIG_TYPES,
    CONFIG_ITEMS
)


FIELD_DATA_WIDTH = 250
PLAIN_TEXT_DATA_HEIGHT = 200


class ConfigUI(QDialog):

    def __init__(
        self, 
        temp_config: LayoutConfig, 
        system_name: str, 
        parent: QWidget = None
    ) -> None:
        super().__init__(parent)
        self.temp_config = temp_config
        self.system_name = system_name
        self.setup_window()

    def select_file(
        self, 
        prompt: str, 
        file_type: str, 
        text_field: QLineEdit
    ) -> None:
        def func() -> None:
            path = QFileDialog.getOpenFileName(
                self, prompt, "", file_type
            )[0]

            if path:
                text_field.setText(path)
        
        return func

    def setup_window(self) -> None:
        self.scroll_widget = QWidget()
        self.scroll_area = QScrollArea()
        self.scroll_area.setWidgetResizable(True)
        self.scroll_layout = QVBoxLayout()
        self.fields = dict()

        current_grid_row = 0
        current_grid_layout: QGridLayout = None
        current_groupbox: QGroupBox = None

        for config_name in CONFIG_ORDER:
            if config_name not in CONFIG_NAMES:
                if current_groupbox is not None:
                    current_groupbox.setLayout(current_grid_layout)
                    self.scroll_layout.addWidget(current_groupbox)
                
                current_grid_row = 0
                current_groupbox = QGroupBox()
                current_groupbox.setTitle(config_name)
                current_grid_layout = QGridLayout()
                continue

            if config_name == SYSTEM_NAME_PLACEHOLDER:
                name_label = QLabel()
                name_label.setText(CONFIG_NAMES[config_name])
                name_data = QLabel()
                name_data.setText(self.system_name)

                current_grid_layout.addWidget(
                    name_label, current_grid_row, 0, 1, 1, Qt.AlignRight
                )
                current_grid_layout.addWidget(
                    name_data, current_grid_row, 1, 1, 2, Qt.AlignLeft
                )
                current_grid_row += 1
                continue

            field_label = QLabel()
            field_label.setText(CONFIG_NAMES[config_name])

            field_type = CONFIG_TYPES[config_name]
            field_data = None
            browse_button = None

            if config_name.startswith(SYSTEM_PREFIX):
                if self.system_name not in self.temp_config.system_map:
                    field_value = None
                else:
                    field_value = self.temp_config.system_map[self.system_name].get(
                        config_name, None
                    )
            else:
                field_value = getattr(self.temp_config, config_name)

            if field_value is None and config_name in CONFIG_ITEMS:
                field_value = CONFIG_ITEMS[config_name]

            if field_type == bool:
                field_data = QCheckBox()
                field_data.setChecked(field_value)

            elif field_type == str:
                field_data = QLineEdit()

                if config_name in CONFIG_FILE_PARAMS:
                    prompt, file_type = CONFIG_FILE_PARAMS[config_name]
                    browse_button = QPushButton("Browse", self)
                    browse_button.clicked.connect(self.select_file(
                        prompt, file_type, field_data
                    ))

                field_data.setText(field_value)

            # Note that all integer configs are percentages (for now)
            # So we're hardcoding the limits
            elif field_type == int:
                field_data = QSpinBox()
                field_data.setRange(5, 10000)
                field_data.setSingleStep(5)
                field_data.setSuffix("%")
                field_data.setValue(field_value or 100)

            if field_data is not None:
                field_data.setMinimumWidth(FIELD_DATA_WIDTH)
                current_grid_layout.addWidget(
                    field_label, current_grid_row, 0, 1, 1, Qt.AlignRight
                )

                data_width = 1 + (browse_button is None) * 1

                current_grid_layout.addWidget(
                    field_data, current_grid_row, 1, 1, data_width, Qt.AlignLeft
                )

                if browse_button is not None:
                    current_grid_layout.addWidget(
                        browse_button, current_grid_row, 2, 1, 1, Qt.AlignLeft
                    )

                self.fields[config_name] = field_data
                current_grid_row += 1

        if current_groupbox is not None:
            current_groupbox.setLayout(current_grid_layout)
            self.scroll_layout.addWidget(current_groupbox)

        self.button_box = QDialogButtonBox(
            (
                QDialogButtonBox.Cancel | 
                QDialogButtonBox.Ok
            ),
            parent=self
        )
        self.button_box.rejected.connect(self.reject)
        self.button_box.accepted.connect(self.save_settings)

        self.scroll_widget.setLayout(self.scroll_layout)
        self.scroll_area.setWidget(self.scroll_widget)

        self.layout = QVBoxLayout()
        self.layout.addWidget(self.scroll_area)
        self.layout.addWidget(self.button_box)
        self.setLayout(self.layout)

    def save_settings(self) -> None:
        for config_name in CONFIG_ITEMS.keys():
            field_type = CONFIG_TYPES[config_name]
            field_data = self.fields[config_name]
            field_value = None

            if field_type == bool:
                field_value = field_data.isChecked()
            elif field_type == int:
                field_value = field_data.value()
            elif field_type == str:
                field_value = field_data.text()

            if self.system_name not in self.temp_config.system_map:
                self.temp_config.system_map[self.system_name] = dict()

            if field_value is not None:
                if config_name.startswith(SYSTEM_PREFIX):
                    self.temp_config.system_map[self.system_name][config_name] = field_value
                else:
                    setattr(self.temp_config, config_name, field_value)
        
        self.accept()



from typing import List, Tuple, Any


CONFIG_ITEMS = {
    "system_svg": "", 
    "system_py": "", 
    "system_scale": 100,
    "force_repaint": False
}

CONFIG_FILE_PARAMS = {
    "system_svg": ("Select Layout SVG", "Vector Graphics (*.svg)"),
    "system_py": ("Select Python Script", "Python Script (*.py)")
}

CONFIG_TYPES = {k: type(v) for k, v in CONFIG_ITEMS.items()}

CONFIG_NAMES = {
    "system_name": "Current System:",
    "system_svg": "Layout SVG",
    "system_py": "Layout Python Script",
    "system_scale": "Layout Scale",
    "force_repaint": "Force Repaint (macOS)"
}

CONFIG_ORDER = [
    "System Settings",
    "system_name",
    "system_svg",
    "system_py",
    "system_scale",

    "Force Repaint (macOS Window Shadow)",
    "force_repaint"
]


SYSTEM_NAME_PLACEHOLDER = "system_name"
SYSTEM_PREFIX = "system_"


class LayoutConfig:
    def __init__(self, system_map: dict = None, values: dict = None) -> None:
        if system_map is None:
            self.system_map = dict()
        else:
            self.system_map = system_map
        
        if values is None:
            values = dict()

        for key, default in CONFIG_ITEMS.items():
            if key in values:
                setattr(self, key, values[key])
            else:
                setattr(self, key, default)
    
    def copy(self) -> "LayoutConfig":
        value_dict = {k: getattr(self, k) for k in CONFIG_ITEMS.keys()}
        return LayoutConfig(self.system_map, value_dict)

    def get_values(self) -> List[Tuple[str, Any]]:
        results = []
        for key in CONFIG_ITEMS.keys():
            if not key.startswith(SYSTEM_PREFIX):
                results.append((key, getattr(self, key)))
        
        return results



from typing import List, Dict
from lxml import etree as ET

from plover_svg_layout_display.qt_utils import load_qt_text
from plover_svg_layout_display.resources_rc import *


SVG_TEMPLATE = (
    "<svg{header}>\n{content}\n</svg>"
)


class SVGParser:
    
    __slots__ = ["group_svgs", "svg_attribs"]

    def load_file(self, path: str) -> List[str]:
        if not path.strip():
            return

        parser = ET.XMLParser(recover=True)
        svg_raw = load_qt_text(path)
        tree = ET.fromstring(svg_raw.encode("utf-8"), parser)

        self.group_svgs: Dict[str, str] = {}
        for child in tree:
            tag = child.tag
            if "}" in tag:
                tag = tag.split("}", 1)[1]

            if tag == "g" and "id" in child.attrib:
                self.group_svgs[child.attrib["id"]] = ET.tostring(child).decode()

        if "<svg" in svg_raw:
            self.svg_attribs = svg_raw.split("<svg", 1)[1].split(">")[0]


    def get_svg_content(
        self, 
        group_ids: List[str]
    ) -> str:
        content = "\n".join(
            self.group_svgs[id] 
            for id in group_ids
            if id in self.group_svgs
        )

        return SVG_TEMPLATE.format(
            header=self.svg_attribs,
            content=content
        )
    
    def get_whole_svg(self) -> str:
        return self.get_svg_content(self.group_svgs.keys())



#!/usr/bin/env python3

from setuptools import setup

from plover_build_utils.setup import BuildPy, BuildUi, Develop

BuildPy.build_dependencies.append('build_ui')
BuildUi.hooks = ['plover_build_utils.pyqt:fix_icons']
Develop.build_dependencies.append('build_py')
CMDCLASS = {
    'build_py': BuildPy,
    'build_ui': BuildUi,
    'develop': Develop,
}

setup(cmdclass=CMDCLASS)



from PyQt5.QtWidgets import QWidget, QSizePolicy
from PyQt5.QtCore import QByteArray, Qt, QRect, QPoint
from PyQt5.QtSvg import QSvgWidget

from typing import List

from plover_svg_layout_display.svg_parser import SVGParser


class LayoutWidget(QSvgWidget):

    def __init__(self, parent: QWidget = None) -> None:
        super().__init__(parent)

        self.setObjectName("layout_widget")
        self.setContextMenuPolicy(Qt.NoContextMenu)
        self.setContentsMargins(0, 0, 0, 0)
        self.setStyleSheet("border:0px; background:transparent;")
        self.renderer().setAspectRatioMode(Qt.KeepAspectRatio)
        self.svg_size = None

        self.svg_parser = SVGParser()
        
    def update_groups(self, group_ids: List[str]) -> None:
        svg_str = self.svg_parser.get_svg_content(group_ids)
        self.load(QByteArray(str.encode(svg_str, "utf-8")))
        self.update()
    
    def load_svg(self, path: str, scale: int) -> None:
        if not path.strip():
            return

        self.svg_parser.load_file(path)
        svg_str = self.svg_parser.get_whole_svg()
        self.load(QByteArray(str.encode(svg_str, "utf-8")))

        scale_ratio = scale / 100
        new_size = self.renderer().defaultSize()
        new_size.scale(
            int(new_size.width() * scale_ratio), 
            int(new_size.height() * scale_ratio), 
            Qt.KeepAspectRatio
        )
        
        self.resize(new_size)
        self.renderer().setViewBox(QRect(QPoint(0, 0), new_size))
        self.svg_size = new_size
        
